/*
** Module   : SectionContent.java
** Abstract : Progress 4GL INI file content class representing one section of data
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20130701 Created initial version.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;

/**
 * A class implementing the INI text file key=value secion lines. Can contain several lines with
 * key values and also comment lines end empty lines in any place of the section.
 *
 * @author    EVL
 */
public class SectionContent
extends Content
{
   /** Map to store key=value pars. */
   private Map<String, String> keyMap = null;
   
   /**
    * Constructs an instance of the data section in INI file.
    *
    * @param name 
    *        The name of the section corresponding to the key=value pairs of the INI file.
    */
   public SectionContent(String name)
   {
      super(name);
      keyMap = new LinkedHashMap<String, String>();
   }
   
   /**
    * Writes the content of the data section to the file via Writer object provided.
    * 
    * @param pw 
    *        The writer will be used to save content to the file.
    */
   public void writeToFile(PrintWriter pw)
   {
      if (keyMap.size() > 0)
      {
         // Write the section name first
         pw.println("[" + name + "]");
         // Key names iterator
         Iterator iterKey = keyMap.keySet().iterator();
         while (iterKey.hasNext())
         {
            // Get next available key=value pair
            String keyName = (String)iterKey.next();
            String keyValue = keyMap.get(keyName);
            // And write them to the file
            // Unrecognized line, write it as it was in INI file
            if (keyName.indexOf(KEY_INRECOGNIZED_MARK) != -1)
            {
               pw.println(keyValue);
            }
            // Usual key-value pair
            else
            {
               pw.println(keyName + "=" + keyValue);
            }
         }
      }
   }
   
   /**
    * Checks if the content of the section is empty or not.
    * 
    * @return <code>true</code> in case of not emptry, <code>false</code> otherwise.
    */
   public boolean notEmpty()
   {
      return keyMap != null && keyMap.size() > 0;
   }
   
   /**
    * Stores the key=value pair into the section internal map.
    * 
    * @param key 
    *        The String key name to store.
    * @param value 
    *        The String value to be associated with the given key name.
    */
   public void store(String key, String value)
   {
      if (keyMap == null)
      {
         keyMap = new LinkedHashMap<String, String>();
      }
      keyMap.put(key, value);
   }

   /**
    * Stores the empty line or comment iside the section content.
    * 
    * @param textLine 
    *        The ignored text line to store.
    */
   public void store(String textLine)
   {
      long i = 0; // Key modifier
      String spaceLineKey = new String(KEY_INRECOGNIZED_MARK+i);
                     
      if (keyMap == null)
      {
         keyMap = new LinkedHashMap<String, String>();
      }

      // Create new unique key name for new line to be added
      while (keyMap.containsKey(spaceLineKey))
      {
         spaceLineKey = new String(KEY_INRECOGNIZED_MARK+(++i));
      }
      // Put "empty" lines into section map as if it is usual key
      keyMap.put(spaceLineKey, textLine);
   }
   
   /**
    * Gets the value associated with the given key inside the section.
    * 
    * @param key 
    *        The String key name of the value to be retrieved.
    * 
    * @return The String value associate with the given key or <code>null</code> if there is no
    *         key=value pair inside the section.
    */
   public String getValueForKey(String key)
   {
      if (keyMap != null)
      {
         return keyMap.get(key);
      }
      
      return null;
   }
}
