/*
** Module   : StanzaIni.java
** Abstract : Progress 4GL compatible Winodws INI files related methods
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20130604 Created initial version.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;

/**
 * A class implementing the INI file based environment methods. The INI file is the plain text
 * file containing the inside the [] braces and key=value pairs within every section.
 *
 * @author    EVL
 */
public class StanzaIni
implements EnvironmentAccessor
{
   /** The name of the INI file that holds the environment variables. */
   private String name = null;
   
   /** Working directory value to search the INI file. */
   private String workingDirectory = null;
   
   /** The flag indicating whether we need to create new file instead of opening existed one. */
   private boolean isNew = false;

   /** The ordered list of INI file content. */
   private List<Content> contList = null;
   
   /**
    * Default constructor creating the internal cache for environment.
    */
   public StanzaIni()
   {
      contList = new LinkedList<Content>();
   }
   
   /**
    * Constructs an instance of the environment with INI file as backend.
    *
    * @param env 
    *        The environment to create or load. The name without extension of the initialization
    *        file.
    * @param directory 
    *        The directory path for existing initialization file or place to create
    *        initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file will be created overwriting the possibly
    *        existed one.
    */
   public StanzaIni(String env, String directory, boolean flagNew)
   {
      this();

      name = env + ".ini";
      if (directory != null)
      {
         workingDirectory = directory;
      }
      isNew = flagNew;
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Initialization file name without extension.
    * @param directory 
    *        The directory path for existing initialization file or place to create
    *        initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file will be created overwriting the possibly
    *        existed one.
    * @param baseKey
    *        When the parameter is specified the searching is performed under the particular key
    *        value. It is possible to specify "INI" as parameter. In this case search or creation
    *        will be dome inside initialization file specified in env parameter.
    */
   public void load(String env, String directory, boolean flagNew, String baseKey)
   {
      // TODO implement this
      if (!isNew)
      {
         loadIniFile();
      }
      else
      {
         createEmptyIniFile();
      }
   }

   /**
    * Specifies environment defaults to use in subsequent windows.
    * 
    * @param env 
    *        The environment to load. If the parameter is null string("") the
    *        default environment becomes the current one.
    */
   public void use(String env)
   {
      // TODO implement this
      return;
   }

   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param key 
    *        The name of the key key to modify or default key if not specified.
    * @param value 
    *        The new value of the key under modification.
    */
   public void setKeyValue(String section, String key, String value)
   {
      // TODO implement this
      // The 4GL does not allow to put extra spaces for sections and key names into INI file
      String sectionAsString = section.trim();
      String keyAsString = key.trim();
      // The 4GL does allow to put extra spaces for value string into INI file
      String valueAsString = value;
      
      SectionContent sectionEntry = getSection(sectionAsString);
      if (sectionEntry == null)
      {
         sectionEntry = new SectionContent(sectionAsString);
      }
      sectionEntry.store(keyAsString, valueAsString);
      contList.add(sectionEntry);
      
      commitIniFile();
   }

   /**
    * Getting the key value from the current environment.
    * 
    * @param section 
    *        The name of the section containing the key to get.
    * @param key 
    *        The name of the key key to get or default key if not specified.
    * 
    * @return The value of the key.
    */
   public String getKeyValue(String section, String key)
   {
      // TODO implement this
      String sectionAsString = section.trim();
      String keyAsString = key.trim();
      String valueAsString = null;
      
      SectionContent sectionEntry = getSection(sectionAsString);
      if (sectionEntry != null)
      {
         valueAsString = sectionEntry.getValueForKey(keyAsString);
      }
      // For missing entries 4GL returns empty string
      if (valueAsString == null)
      {
         return new String("");
      }
      else
      {
         // For statement GET-KEY-VALUE 4GL does not allow to return extra spaces unlike to
         // PUT-KEY-VALUE. Pretty strange behavior but was tested on real Windows 4GL system.
         return valueAsString.trim();
      }
   }
   
   /**
    * Getting the object that holds the section from the previously formed environment.
    * 
    * @param name 
    *        The name of the section to look for.
    * 
    * @return The SectionContent object or null if the section was not found.
    */
   private SectionContent getSection(String name)
   {
      Iterator contentIter = contList.iterator();
      while (contentIter.hasNext())
      {
         // Walk through the content object and check for name match
         Content contentEntry = (Content)contentIter.next();
         if (contentEntry.withName(name))
         {
            return (SectionContent)contentEntry;
         }
      }
      
      return null;
   }

   /**
    * Unloads specifies environment from the current one.
    * 
    * @param env 
    *        The environment to unload. For INI file it is initialization file name without
    *        extension. The environment to unload should be previously loaded by load() method.
    */
   public void unload(String env)
   {
      // TODO implement this
      return;
   }

   /**
    * Reconstructs full initial file name with possible path.
    */
   private String getFullIniFileName()
   {
      StringBuilder fullName = new StringBuilder();
      
      // We have non-empty directory value, add it first 
      if (workingDirectory != null && workingDirectory.length() > 0)
      {
         fullName.append(workingDirectory);
         fullName.append(EnvironmentOps.getLegacyFileSeparator());
      }
      // And we have the non-empty file name, add it
      if (name != null && name.length() > 0)
      {
         fullName.append(name);
      }
      // If something was added, convert to String
      if (fullName.length() > 0)
      {
         return fullName.toString();
      }
      else // Otherwise return null
      {
         return null;
      }
   }
   
   /**
    * Reads the current environment set from appropriate INI file in file system.
    */
   private void loadIniFile()
   {
      String fileName = getFullIniFileName();
      
      if (fileName != null && fileName.length() > 0)
      {
         BufferedReader br = null;
         
         try
         {
            // Open file stream for reading
            br = new BufferedReader(new FileReader(fileName));

            String sectionName = null;
            Content contentEntry = null;
            
            String nextLine = br.readLine();
            // Read until EOF(nextLine == null)
            while (nextLine != null)
            {
               // Next INI file section should start here
               if (nextLine.length() > 2 && nextLine.charAt(0) == '[' &&
                   nextLine.charAt(nextLine.length() - 1) == ']')
               {
                  // The previously constructed section should be saved before starting to
                  // process the next section
                  if (sectionName != null && contentEntry != null && contentEntry.notEmpty())
                  {
                     contList.add(contentEntry);
                  }
                  // The new section name candidate
                  sectionName = new String(nextLine.substring(1,nextLine.length() - 1));
                  // Purge the previous one if it was existed
                  if (contentEntry != null)
                  {
                     contentEntry = null;
                  }
               }
               // Section processing
               else
               {
                  // Key-value candidate, get the index of separator first
                  int separatorNdx = nextLine.indexOf('=');
                  // Something should exist before and after separator and
                  // line is not a comment
                  if (separatorNdx > 0 && nextLine.length() > separatorNdx + 1 &&
                      nextLine.trim().indexOf(';') != 0)
                  {
                     if (contentEntry == null)
                     {
                        contentEntry = new SectionContent(sectionName);
                     }
                     // Construct key-value pair from line string
                     // Key name should be trimmed, need to check other cases
                     String keyName = new String(nextLine.substring(0,separatorNdx).trim());
                     // The value should remain as it is in INI file on reading stage.
                     String keyValue = new String(nextLine.substring(separatorNdx + 1));
                     // Store the next key-value pair into section content
                     ((SectionContent)contentEntry).store(keyName, keyValue);
                  }
                  // Here we add artificial meaningless data including empty lines between
                  // sections
                  else
                  {
                     // There was no sections at this time, create simple line 
                     if (contentEntry == null)
                     {
                        contentEntry = new LineContent(nextLine);
                        // Put the simple line into ordered list
                        contList.add(contentEntry);
                        // And mark this is not a section start
                        contentEntry = null;
                     }
                     // The line is a part of the section already created 
                     else
                     {
                        // Store the simple line inside the section
                        ((SectionContent)contentEntry).store(nextLine);
                     }
                  }
               }
               
               // Read the next line from file
               nextLine = br.readLine();
               if (nextLine == null && sectionName != null &&
                   contentEntry != null && contentEntry.notEmpty())
               {
                  // Possible last section in the file
                  contList.add(contentEntry);
               }
            }
            br.close();
         }
         catch (IOException ioex)
         {
         }
      }
   }
   
   /**
    * Creates the empty INI file for current environment set in file system.
    */
   private void createEmptyIniFile()
   {
      String fileName = getFullIniFileName();
      
      if (fileName != null && fileName.length() > 0)
      {
         PrintWriter pw = null;
         
         try
         {
            // Open file for writing
            pw = new PrintWriter(new FileWriter(fileName, false));
         }
         catch (IOException ioex)
         {
         }
         finally
         {
            // Do nothing and close it
            if (pw != null)
            {
               pw.close();
            }
         }
      }
   }
   
   /**
    * Writes the current environment set to the appropriate INI file in file system.
    */
   private void commitIniFile()
   {
      String fileName = getFullIniFileName();
      
      if (fileName != null && fileName.length() > 0)
      {
         PrintWriter pw = null;
         
         try
         {
            // Open file for writing
            pw = new PrintWriter(new FileWriter(fileName, false));
            // Sections iterator
            Iterator iterEntry = contList.iterator();
            while (iterEntry.hasNext())
            {
               // Get next available content entry, section or ignoring line
               Content contEntry = (Content)iterEntry.next();
               // If entry is not empty write the content to file
               if (contEntry.notEmpty())
               {
                  contEntry.writeToFile(pw);
               }
            }
            // Commit the changes
            pw.flush();
         }
         catch (IOException ioex)
         {
         }
         finally
         {
            if (pw != null)
            {
               // And finally close the file
               pw.close();
            }
         }
      }
   }
}
